import logging
import os
from itertools import groupby
from operator import attrgetter

import meshio
import numpy as np

from ada.config import Settings as _Settings
from ada.fem import ElemShapes


def meshio_to_fem(
    assembly,
    name,
    scratch_dir=None,
    metadata=None,
    execute=False,
    run_ext=False,
    cpus=2,
    gpus=None,
    overwrite=False,
    exit_on_complete=True,
):
    """

    :param assembly:
    :param name:
    :param scratch_dir:
    :param metadata:
    :param execute:
    :param run_ext:
    :param cpus:
    :param gpus:
    :param overwrite:
    :param exit_on_complete:

    :type assembly: ada.Assembly
    :return:
    """
    if scratch_dir is None:
        scratch_dir = _Settings.scratch_dir

    mesh_format = metadata["fem_format"]
    analysis_dir = os.path.join(scratch_dir, name)

    os.makedirs(analysis_dir, exist_ok=True)

    for p in assembly.get_all_parts_in_assembly(include_self=True):
        mesh = fem_to_meshio(p.fem)
        if mesh is None:
            continue
        prefix_mapper = dict(abaqus="inp")
        if mesh_format in ["abaqus"]:
            prefix = prefix_mapper[mesh_format]
        else:
            prefix = mesh_format

        part_file = os.path.join(analysis_dir, f"{p.name}_bulk/{p.name}.{prefix}")
        os.makedirs(os.path.dirname(part_file), exist_ok=True)
        meshio.write(
            part_file,  # str, os.PathLike, or buffer/ open file
            mesh,
            file_format=mesh_format,
        )
        print(f'Exported "{mesh_format}" using meshio to "{analysis_dir}"')


def fem_to_meshio(fem):
    """

    :param fem: Ada FEM object
    :type fem: ada.fem.FEM
    :return: Meshio MESH object
    :rtype: meshio.Mesh
    """
    from meshio.abaqus._abaqus import abaqus_to_meshio_type

    if len(fem.nodes) == 0:
        logging.error(f"Attempt to convert empty FEM mesh for ({fem.name}) aborted")
        return None

    # Points
    points = np.zeros((int(fem.nodes.max_nid + 1), 3))

    def pmap(n):
        points[int(n.id - 1)] = n.p

    list(map(pmap, fem.nodes))

    # Elements

    def get_node_ids_from_element(el_):
        return [int(n.id - 1) for n in el_.nodes]

    cells = []
    for group, elements in groupby(fem.elements, key=attrgetter("type")):
        if group in ElemShapes.masses + ElemShapes.springs:
            logging.error("NotImplemented: Skipping Mass or Spring Elements")
            continue
        med_el = abaqus_to_meshio_type[group]
        elements = list(elements)
        el_mapped = np.array(list(map(get_node_ids_from_element, elements)))
        el_long = np.zeros((int(fem.elements.max_el_id + 1), len(el_mapped[0])))
        for el in elements:
            el_long[el.id] = get_node_ids_from_element(el)

        cells.append((med_el, el_mapped))

    cell_sets = dict()
    for set_name, elset in fem.sets.elements.items():
        cell_sets[set_name] = np.array([[el.id for el in elset.members]], dtype="int32")

    point_sets = dict()
    for set_name, nset in fem.sets.nodes.items():
        point_sets[set_name] = np.array([[el.id for el in nset.members]], dtype="int32")

    mesh = meshio.Mesh(points, cells, point_sets=point_sets, cell_sets=cell_sets)
    return mesh



import logging

import ifcopenshell
import ifcopenshell.geom
import ifcopenshell.util.element
import numpy as np
from ifcopenshell.util.unit import get_prefix_multiplier

import ada.core.constants as ifco
from ada.core.utils import Counter, create_guid, get_list_of_files, roundoff

name_gen = Counter(1, "IfcEl")


def create_ifcaxis2placement(ifcfile, origin=ifco.O, loc_z=ifco.Z, loc_x=ifco.X):
    """
    Creates an IfcAxis2Placement3D from Location, Axis and RefDirection specified as Python tuples

    :param ifcfile:
    :param origin:
    :param loc_z:
    :param loc_x:
    :return:
    """

    ifc_origin = ifcfile.createIfcCartesianPoint(to_real(origin))
    ifc_loc_z = ifcfile.createIfcDirection(to_real(loc_z))
    ifc_loc_x = ifcfile.createIfcDirection(to_real(loc_x))
    axis2placement = ifcfile.createIfcAxis2Placement3D(ifc_origin, ifc_loc_z, ifc_loc_x)
    return axis2placement


def create_ifclocalplacement(ifcfile, origin=ifco.O, loc_z=ifco.Z, loc_x=ifco.X, relative_to=None):
    """
    Creates an IfcLocalPlacement from Location, Axis and RefDirection,
    specified as Python tuples, and relative placement

    :param ifcfile:
    :param origin:
    :param loc_z:
    :param loc_x:
    :param relative_to:
    :return: IFC local placement
    """

    axis2placement = create_ifcaxis2placement(ifcfile, origin, loc_z, loc_x)
    ifclocalplacement2 = ifcfile.createIfcLocalPlacement(relative_to, axis2placement)
    return ifclocalplacement2


def generate_tpl_ifc_file(file_name, project, organization, creator, schema, units):
    """

    :param file_name:
    :param project:
    :param organization:
    :param creator:
    :param schema:
    :param units:
    :return:
    """

    import time

    from ada.core.ifc_template import tpl_create

    timestamp = time.time()
    timestring = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(timestamp))
    application, application_version = "IfcOpenShell", "0.6"
    project_globalid = create_guid()
    if units == "m":
        units_str = "$,.METRE."
    elif units == "mm":
        units_str = ".MILLI.,.METRE."
    else:
        raise ValueError(f'Unrecognized unit prefix "{units}"')
    ifc_file = tpl_create(
        file_name + ".ifc",
        timestring,
        organization,
        creator,
        schema,
        application_version,
        int(timestamp),
        application,
        project_globalid,
        project,
        units_str,
    )

    return ifc_file


def create_ifcpolyline(ifcfile, point_list):
    """
    Creates an IfcPolyLine from a list of points, specified as Python tuples

    :param ifcfile:
    :param point_list:
    :return:
    """
    ifcpts = []
    for p_in in point_list:
        p = [float(x) for x in p_in]
        point = ifcfile.createIfcCartesianPoint(p)
        ifcpts.append(point)
    polyline = ifcfile.createIfcPolyLine(ifcpts)
    return polyline


def create_ifcindexpolyline(ifcfile, points, seg_index):
    """
    Assumes a point list whereas all points that are to be used for creating arc-segments will have 4 values
    (x,y,z,r) instead of 3 (x,y,z)

    #https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/link/annex-e.htm

    :param ifcfile:
    :param points:
    :param seg_index:
    :return:
    """
    ifc_segments = []
    for seg_ind in seg_index:
        if len(seg_ind) == 2:
            ifc_segments.append(ifcfile.createIfcLineIndex(seg_ind))
        elif len(seg_ind) == 3:
            ifc_segments.append(ifcfile.createIfcArcIndex(seg_ind))
        else:
            raise ValueError("Unrecognized number of values")

    # TODO: Investigate using 2DLists instead is it could reduce complexity?
    # ifc_point_list = ifcfile.createIfcCartesianPointList2D(points)

    ifc_point_list = ifcfile.createIfcCartesianPointList3D(points)
    segindex = ifcfile.createIfcIndexedPolyCurve(ifc_point_list, ifc_segments)
    return segindex


def create_ifcindexpolyline2d(ifcfile, points2d, seg_index):
    """
    Assumes a point list whereas all points that are to be used for creating arc-segments will have 4 values
    (x,y,z,r) instead of 3 (x,y,z)

    #https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/link/annex-e.htm

    :param ifcfile:
    :param points2d:
    :param seg_index:
    :return:
    """
    ifc_segments = []
    for seg_ind in seg_index:
        if len(seg_ind) == 2:
            ifc_segments.append(ifcfile.createIfcLineIndex(seg_ind))
        elif len(seg_ind) == 3:
            ifc_segments.append(ifcfile.createIfcArcIndex(seg_ind))
        else:
            raise ValueError("Unrecognized number of values")

    # TODO: Investigate using 2DLists instead is it could reduce complexity?
    # ifc_point_list = ifcfile.createIfcCartesianPointList2D(points)

    ifc_point_list = ifcfile.createIfcCartesianPointList2D(points2d)
    segindex = ifcfile.createIfcIndexedPolyCurve(ifc_point_list, ifc_segments)
    return segindex


def create_ifcrevolveareasolid(ifc_file, profile, ifcaxis2placement, origin, revolve_axis, revolve_angle):
    """
    Creates an IfcExtrudedAreaSolid from a list of points, specified as Python tuples
    :param ifc_file:
    :param profile:
    :param ifcaxis2placement:
    :param origin:
    :param revolve_axis:
    :param revolve_angle:
    :return:
    """
    ifcorigin = ifc_file.createIfcCartesianPoint(origin)
    ifcaxis1dir = ifc_file.createIfcAxis1Placement(ifcorigin, ifc_file.createIfcDirection(revolve_axis))

    ifcextrudedareasolid = ifc_file.createIfcRevolvedAreaSolid(profile, ifcaxis2placement, ifcaxis1dir, revolve_angle)
    return ifcextrudedareasolid


def create_IfcFixedReferenceSweptAreaSolid(
    ifc_file, curve, profile, ifcaxis2placement, start_param, end_param, fixed_ref
):
    """

    :param ifc_file:
    :param profile:
    :param ifcaxis2placement:
    :param curve:
    :param start_param:
    :param end_param:
    :param fixed_ref:
    :return:
    """
    ifcextrudedareasolid = ifc_file.createIfcFixedReferenceSweptAreaSolid(
        profile, ifcaxis2placement, curve, start_param, end_param, fixed_ref
    )
    return ifcextrudedareasolid


def create_ifcextrudedareasolid(ifc_file, profile, ifcaxis2placement, extrude_dir, extrusion):
    """
    Creates an IfcExtrudedAreaSolid from a list of points, specified as Python tuples

    :param profile:
    :param ifcaxis2placement:
    :param extrude_dir:
    :param extrusion:
    :return:
    """

    ifcdir = ifc_file.createIfcDirection(extrude_dir)
    ifcextrudedareasolid = ifc_file.createIfcExtrudedAreaSolid(profile, ifcaxis2placement, ifcdir, extrusion)
    return ifcextrudedareasolid


def create_ifcrightcylinder(ifc_file, ifcaxis2placement, height, radius):
    """

    :param ifc_file:
    :param ifcaxis2placement:
    :param height:
    :param radius:
    :return:
    """
    ifcextrudedareasolid = ifc_file.createIfcRightCircularCylinder(ifcaxis2placement, height, radius)
    return ifcextrudedareasolid


def create_property_set(name, ifc_file, metadata_props):
    owner_history = ifc_file.by_type("IfcOwnerHistory")[0]
    properties = []

    def ifc_value(v_):
        return ifc_file.create_entity("IfcText", str(v_))

    def to_str(in_enum):
        return (
            f"{in_enum}".replace("(", "")
            .replace(")", "")
            .replace(" ", "")
            .replace(",", ";")
            .replace("[", "")
            .replace("]", "")
        )

    for key, value in metadata_props.items():
        if type(value) in (tuple, list):
            if type(value[0]) in (list, tuple, np.ndarray):
                for i, v in enumerate(value):
                    if type(v) is np.ndarray:
                        v = [roundoff(x) for x in v]
                    properties.append(
                        ifc_file.create_entity(
                            "IfcPropertySingleValue",
                            Name=f"{key}_{i}",
                            NominalValue=ifc_value(to_str(v)),
                        )
                    )
            else:
                properties.append(
                    ifc_file.create_entity(
                        "IfcPropertySingleValue",
                        Name=key,
                        NominalValue=ifc_value(to_str(value)),
                    )
                )
        else:
            properties.append(
                ifc_file.create_entity(
                    "IfcPropertySingleValue",
                    Name=key,
                    NominalValue=ifc_value(value),
                )
            )

    atts = {
        "GlobalId": ifcopenshell.guid.new(),
        "OwnerHistory": owner_history,
        "Name": name,
        "HasProperties": properties,
    }

    return ifc_file.create_entity("IfcPropertySet", **atts)


def add_properties_to_elem(name, ifc_file, ifc_elem, elem_props):
    """

    :param ifc_file:
    :param ifc_elem:
    :param elem_props:
    :return:
    """
    owner_history = ifc_file.by_type("IfcOwnerHistory")[0]
    props = create_property_set(name, ifc_file, elem_props)
    ifc_file.createIfcRelDefinesByProperties(
        create_guid(),
        owner_history,
        "Properties",
        None,
        [ifc_elem],
        props,
    )


def to_real(v):
    """

    :param v:
    :return:
    """
    if type(v) is tuple:
        return [float(x) for x in v]
    elif type(v) is list:
        if type(v[0]) is float:
            return v
        else:
            return [float(x) for x in v]
    else:
        return v.astype(float).tolist()


def getIfcPropertySets(ifcelem):
    """Returns a dictionary of {pset_id:[prop_id, prop_id...]} for an IFC object"""
    props = dict()
    # get psets for this pid
    for definition in ifcelem.IsDefinedBy:
        # To support IFC2X3, we need to filter our results.
        if definition.is_a("IfcRelDefinesByProperties"):
            property_set = definition.RelatingPropertyDefinition
            pset_name = property_set.Name.split(":")[0].strip()
            props[pset_name] = dict()
            if property_set.is_a("IfcElementQuantity"):
                continue
            for prop in property_set.HasProperties:
                if prop.is_a("IfcPropertySingleValue"):
                    props[pset_name][prop.Name] = prop.NominalValue.wrappedValue
            # Returning first instance of RelDefines
            # return props (Why?)
    return props


def get_parent(instance):
    """

    :param instance:
    :return:
    """
    if instance.is_a("IfcOpeningElement"):
        return instance.VoidsElements[0].RelatingBuildingElement
    if instance.is_a("IfcElement"):
        fills = instance.FillsVoids
        if len(fills):
            return fills[0].RelatingOpeningElement
        containments = instance.ContainedInStructure
        if len(containments):
            return containments[0].RelatingStructure
    if instance.is_a("IfcObjectDefinition"):
        decompositions = instance.Decomposes
        if len(decompositions):
            return decompositions[0].RelatingObject


def get_association(ifc_elem):
    """

    :param ifc_elem:
    :return:
    """
    c = None
    for association in ifc_elem.HasAssociations:
        if association.is_a("IfcRelAssociatesMaterial"):
            material = association.RelatingMaterial
            if material.is_a("IfcMaterialProfileSet"):
                # For now, we only deal with a single profile
                c = material.MaterialProfiles[0]
            if material.is_a("IfcMaterialProfileSetUsage"):
                c = material.ForProfileSet.MaterialProfiles[0]
            if material.is_a("IfcRelAssociatesMaterial"):
                c = material.RelatingMaterial
            if material.is_a("IfcMaterial"):
                c = material
    if c is None:
        raise ValueError(f'IfcElem "{ifc_elem.Name}" lacks associated Material properties')

    return c


def get_name(ifc_elem):
    """

    :param ifc_elem:
    :return:
    """
    props = getIfcPropertySets(ifc_elem)
    product_name = ifc_elem.Name
    if hasattr(props, "NAME") and product_name is None:
        name = props["NAME"]
    else:
        name = product_name
    if name is None:
        name = next(name_gen)
    return name


def get_representation(ifc_elem, settings):
    """

    :param ifc_elem:
    :param settings:
    :return:
    """
    pdct_shape = ifcopenshell.geom.create_shape(settings, inst=ifc_elem)

    if pdct_shape is None:
        print(f'Unable to import geometry for ifc element "{ifc_elem}"')
        return pdct_shape, None, None

    geom = get_geom(ifc_elem, settings)
    r, g, b, alpha = pdct_shape.styles[0]  # the shape color
    return geom, (r, g, b), alpha


def get_geom(ifc_elem, settings):
    """

    :param ifc_elem:
    :param settings:
    :return:
    """
    from ifcopenshell.geom.occ_utils import shape_tuple
    from OCC.Core import BRepTools
    from OCC.Core.TopoDS import TopoDS_Compound

    try:
        pdct_shape = ifcopenshell.geom.create_shape(settings, inst=ifc_elem)
    except RuntimeError:
        print(f'unable to parse ifc_elem "{ifc_elem}"')
        return

    if type(pdct_shape) is shape_tuple:
        shape = pdct_shape[1]
    else:
        shape = pdct_shape.solid

    if type(shape) is not TopoDS_Compound:
        brep_data = pdct_shape.solid.brep_data
        ss = BRepTools.BRepTools_ShapeSet()
        ss.ReadFromString(brep_data)
        nb_shapes = ss.NbShapes()
        occ_shape = ss.Shape(nb_shapes)
    else:
        occ_shape = shape
    return occ_shape


def import_indexedpolycurve(ipoly, normal, xdir, origin):
    """

    :param ipoly: IFC element
    :param normal:
    :param xdir:
    :param origin:
    :return:
    """
    from ada import ArcSegment, LineSegment
    from ada.core.utils import global_2_local_nodes, segments_to_local_points

    ydir = np.cross(normal, xdir)
    nodes3d = [p for p in ipoly.Points.CoordList]
    nodes2d = global_2_local_nodes([xdir, ydir], origin, nodes3d)
    nodes2d = [np.array([n[0], n[1], 0.0]) for n in nodes2d]
    seg_list = []
    for i, seg in enumerate(ipoly.Segments):
        if seg.is_a("IfcLineIndex"):
            v = seg.wrappedValue
            p1 = nodes2d[v[0] - 1]
            p2 = nodes2d[v[1] - 1]
            seg_list.append(LineSegment(p1=p1, p2=p2))
        elif seg.is_a("IfcArcIndex"):
            v = seg.wrappedValue
            p1 = nodes2d[v[0] - 1]
            p2 = nodes2d[v[1] - 1]
            p3 = nodes2d[v[2] - 1]
            seg_list.append(ArcSegment(p1, p3, midpoint=p2))
        else:
            raise ValueError("Unrecognized type")

    local_points = [(roundoff(x[0]), roundoff(x[1])) for x in segments_to_local_points(seg_list)]
    return local_points


def import_polycurve(poly, normal, xdir):
    from ada.core.utils import global_2_local_nodes

    ydir = np.cross(normal, xdir)
    nodes3d = [p for p in poly.Points]
    nodes2d = global_2_local_nodes([xdir, ydir], (0, 0, 0), nodes3d)

    return nodes2d


def add_negative_extrusion(f, origin, loc_z, loc_x, depth, points, parent):
    """

    :param f:
    :param origin:
    :param loc_z:
    :param loc_x:
    :param depth:
    :param points:
    :param parent:
    :return:
    """

    context = f.by_type("IfcGeometricRepresentationContext")[0]
    owner_history = f.by_type("IfcOwnerHistory")[0]

    # Create and associate an opening for the window in the wall
    opening_placement = create_ifclocalplacement(f, origin, loc_z, loc_x, parent.ObjectPlacement)
    opening_axis_placement = create_ifcaxis2placement(f, origin, loc_z, loc_x)
    polyline = create_ifcpolyline(f, points)
    ifcclosedprofile = f.createIfcArbitraryClosedProfileDef("AREA", None, polyline)

    opening_solid = create_ifcextrudedareasolid(f, ifcclosedprofile, opening_axis_placement, loc_z, depth)
    opening_representation = f.createIfcShapeRepresentation(context, "Body", "SweptSolid", [opening_solid])
    opening_shape = f.createIfcProductDefinitionShape(None, None, [opening_representation])
    opening_element = f.createIfcOpeningElement(
        create_guid(),
        owner_history,
        "Opening",
        "Door opening",
        None,
        opening_placement,
        opening_shape,
        None,
    )
    f.createIfcRelVoidsElement(create_guid(), owner_history, None, None, parent, opening_element)

    return opening_element


def add_colour(f, ifc_body, name, colour):
    """

    :param f:
    :param ifc_body:
    :param name:
    :param colour:
    :return:
    """
    colour = f.createIfcColourRgb(name, colour[0], colour[1], colour[2])
    surfaceStyleShading = f.createIfcSurfaceStyleShading()
    surfaceStyleShading.SurfaceColour = colour
    surfaceStyle = f.createIfcSurfaceStyle(colour.Name, "BOTH", (surfaceStyleShading,))
    presStyleAssign = f.createIfcPresentationStyleAssignment((surfaceStyle,))
    f.createIfcStyledItem(ifc_body, (presStyleAssign,), colour.Name)


def calculate_unit_scale(file):
    units = file.by_type("IfcUnitAssignment")[0]
    unit_scale = 1
    for unit in units.Units:
        if not hasattr(unit, "UnitType") or unit.UnitType != "LENGTHUNIT":
            continue
        while unit.is_a("IfcConversionBasedUnit"):
            unit_scale *= unit.ConversionFactor.ValueComponent.wrappedValue
            unit = unit.ConversionFactor.UnitComponent
        if unit.is_a("IfcSIUnit"):
            unit_scale *= get_prefix_multiplier(unit.Prefix)
    return unit_scale


def scale_ifc_file_object(ifc_file, scale_factor):
    """
    Scale length factor to meter

    :param ifc_file:
    :return:
    """

    s = ifcopenshell.ifcopenshell_wrapper.schema_by_name("IFC4")
    classes_to_modify = {}
    for d in s.declarations():
        if not hasattr(d, "all_attributes") or "IfcLength" not in str(d.all_attributes()):
            continue
        attributes_to_modify = []
        for attribute in d.all_attributes():
            if "IfcLength" in str(attribute):
                attributes_to_modify.append(attribute.name())
        classes_to_modify[d.name()] = attributes_to_modify

    def scale_all(obj, sf):
        def serialize(obj_):
            """Recursively walk object's hierarchy."""
            if isinstance(obj_, (int, float)):
                return obj_ * sf
            elif isinstance(obj_, list):
                return [serialize(item) for item in obj_]
            elif isinstance(obj_, tuple):
                return tuple(serialize([item for item in obj_]))
            else:
                try:
                    if obj_.is_a("IfcLengthMeasure") is True:
                        obj_.wrappedValue = obj_.wrappedValue * sf
                        return obj_
                    elif obj_.is_a("IfcReal") is True:
                        obj_.wrappedValue = obj_.wrappedValue * sf
                        return obj_
                    elif obj_.is_a("IfcInteger") is True:
                        obj_.wrappedValue = int(obj_.wrappedValue * sf)
                        return obj_
                    elif obj_.is_a("IfcPlaneAngleMeasure") is True:
                        return obj_
                    elif obj_.is_a("IfcPressureMeasure") or obj_.is_a("IfcModulusOfElasticityMeasure"):
                        # sf is a length unit.
                        conv_unit = 1 / sf ** 2
                        obj_.wrappedValue = obj_.wrappedValue * conv_unit
                        return obj_
                    elif obj_.is_a("IfcMassDensityMeasure"):
                        conv_unit = 1 / sf ** 3
                        obj_.wrappedValue = obj_.wrappedValue * conv_unit
                        return obj_
                    # Unit-less
                    elif obj_.is_a("IfcText") is True or obj_.is_a("IfcPositiveRatioMeasure") is True:
                        return obj_
                    elif obj_.is_a("IfcThermalExpansionCoefficientMeasure") or obj_.is_a(
                        "IfcSpecificHeatCapacityMeasure"
                    ):
                        return obj_
                    elif obj_.is_a("IfcLogical") is True:
                        return obj_
                except Exception as er:
                    raise ValueError(f"Error {er}")

                raise ValueError(f'Unknown entity "{type(obj_)}", "{obj_}"')

        return serialize(obj)

    for ifc_class, attributes in classes_to_modify.items():
        for element in ifc_file.by_type(ifc_class):
            for attribute in attributes:

                old_val = getattr(element, attribute)
                if old_val is None:
                    continue
                try:
                    setattr(element, attribute, scale_all(old_val, scale_factor))
                except Exception as e:
                    raise ValueError(e)
                # new_val = getattr(element, attribute)
    return ifc_file


def add_fem(ifc_file, part):
    """
    :param ifc_file:
    :param part:
    :type part: ada.Part
    :return:
    """
    from itertools import groupby
    from operator import attrgetter

    from ada.fem import ElemShapes

    ifc_file.createIfcStructuralAnalysisModel(
        create_guid(),
        ifc_file.by_type("IfcOwnerHistory")[0],
        "Structural Analysis",
        "Example analysis",
        ".NOTDEFINED.",
        "LOADING_3D",
    )
    fem = part.fem

    for el_type, elements in groupby(fem.elements, key=attrgetter("type")):

        if el_type in ElemShapes.beam:
            for elem in elements:
                # TODO: add_fem_elem_beam(elem, parent)
                pass


def merge_existing(original_file, source_file, new_file):
    source = ifcopenshell.open(source_file)
    f = ifcopenshell.open(original_file)
    original_project = f.by_type("IfcProject")[0]
    merged_project = f.add(source.by_type("IfcProject")[0])
    for element in source.by_type("IfcRoot"):
        f.add(element)
    for inverse in f.get_inverse(merged_project):
        ifcopenshell.util.element.replace_attribute(inverse, merged_project, original_project)
    f.remove(merged_project)
    f.write(str(new_file))


def patch(f, source_file):
    """

    :param f:
    :param source_file:
    :return:
    """
    source = ifcopenshell.open(source_file)
    original_project = f.by_type("IfcProject")[0]
    merged_project = f.add(source.by_type("IfcProject")[0])
    for element in source.by_type("IfcRoot"):
        f.add(element)
    for inverse in f.get_inverse(merged_project):
        ifcopenshell.util.element.replace_attribute(inverse, merged_project, original_project)
    f.remove(merged_project)


def merge_ifc_files(parent_dir, output_file_name, clean_files=False, include_elements=None, exclude_elements=None):
    """

    :param parent_dir:
    :param output_file_name:
    :param clean_files:
    :param include_elements:
    :param exclude_elements:
    :return:
    """
    import pathlib
    import time

    parent_dir = pathlib.Path(parent_dir)
    files = get_list_of_files(parent_dir, ".ifc")
    f = ifcopenshell.open(files[0])
    for i, fp in enumerate(files[1:]):
        checkpoint = time.time()
        print(f'merging products ({i+1} of {len(files)-1}) from "{fp}"')
        fn = ifcopenshell.open(fp)
        print(f"file opened in {time.time() - checkpoint:.2f} seconds")
        checkpoint = time.time()
        for product in fn.by_type("IfcProduct"):
            f.add(product)
        print(f"Products added in {time.time() - checkpoint:.2f} seconds")
        del fn
    out_file_name = str((parent_dir / output_file_name).with_suffix(".ifc"))
    print(f'Writing file "{out_file_name}"')
    checkpoint = time.time()
    f.write(out_file_name)
    print(f"File written in {time.time() - checkpoint:.2f} seconds")


def convert_bm_jusl_to_ifc(bm):
    """
    IfcCardinalPointReference


    1.      bottom left
    2.      bottom centre
    3.      bottom right
    4.      mid-depth left
    5.      mid-depth centre
    6.      mid-depth right
    7.      top left
    8.      top centre
    9.      top right
    10.     geometric centroid
    11.     bottom in line with the geometric centroid
    12.     left in line with the geometric centroid
    13.     right in line with the geometric centroid
    14.     top in line with the geometric centroid
    15.     shear centre
    16.     bottom in line with the shear centre
    17.     left in line with the shear centre
    18.     right in line with the shear centre
    19.     top in line with the shear centre

    https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/schema/ifcmaterialresource/lexical/ifccardinalpointreference.htm

    :param bm:
    :type bm: ada.Beam
    :return:
    """
    jusl = bm.jusl
    jusl_map = dict(NA=5, TOP=8)

    jusl_val = jusl_map.get(jusl, None)

    if jusl_val is None:
        if jusl != "NA":
            logging.error(f'Unknown JUSL value "{jusl}". Using NA')
        return 5

    return jusl_val



import logging

import numpy as np

from ada.core.containers import Materials
from ada.core.utils import get_current_user
from ada.fem import FemSection, Load
from ada.fem.io.utils import _folder_prep


def to_fem(
    assembly,
    name,
    scratch_dir=None,
    metadata=None,
    execute=False,
    run_ext=False,
    cpus=2,
    gpus=None,
    overwrite=False,
    exit_on_complete=False,
):
    if metadata is None:
        metadata = dict()
    if "control_file" not in metadata.keys():
        metadata["control_file"] = None

    a = SesamWriter(assembly)
    a.write(name, scratch_dir, metadata, overwrite=overwrite)


class SesamWriter:
    """
    `Sesam <https://www.dnv.com/software/products/sesam-products.html>`_ is a brand of services from DNV covering
    FEM and Hydrodynamics for maritime applications.

    """

    def __init__(self, assembly):
        parts = list(filter(lambda x: len(x.fem.nodes) > 0, assembly.get_all_subparts()))
        if len(parts) != 1:
            raise ValueError("Sesam writer currently only works for a single part")
        part = parts[0]
        from operator import attrgetter

        self._gnodes = sorted(part.fem.nodes, key=attrgetter("id"))
        self._gloads = part.fem.steps[0].loads if len(part.fem.steps) > 0 else []
        self._gelements = part.fem.elements
        self._gsections = part.fem.sections
        self._gmaterials = Materials()
        for fsec in self._gsections:
            self._gmaterials.add(fsec.material)
        self._gelsets = part.fem.elsets
        self._gnsets = part.fem.nsets
        self._gmass = part.fem.masses
        self._gbcs = part.fem.bcs
        self._gnonstru = []
        self._geccen = []
        self._thick_map = None

    def write(
        self,
        name,
        scratch_dir=None,
        description=None,
        execute=False,
        run_ext=False,
        cpus=2,
        gpus=None,
        overwrite=False,
    ):
        analysis_dir = _folder_prep(scratch_dir, name, overwrite)
        import datetime

        now = datetime.datetime.now()
        date_str = now.strftime("%d-%b-%Y")
        clock_str = now.strftime("%H:%M:%S")
        user = get_current_user()

        units = "UNITS     5.00000000E+00  1.00000000E+00  1.00000000E+00  1.00000000E+00\n          1.00000000E+00\n"

        with open((analysis_dir / "T100").with_suffix(".FEM"), "w") as d:
            d.write(
                f"""IDENT     1.00000000E+00  1.00000000E+02  3.00000000E+00  0.00000000E+00
DATE      1.00000000E+00  0.00000000E+00  4.00000000E+00  7.20000000E+01
DATE:     {date_str}         TIME:          {clock_str}
PROGRAM:  ADA python          VERSION:       Not Applicable
COMPUTER: X86 Windows         INSTALLATION:
USER:     {user}            ACCOUNT:     \n"""
            )
            d.write(units)
            d.write(self._materials_str)
            d.write(self._sections_str)
            d.write(self._nodes_str)
            d.write(self._mass_str)
            d.write(self._bc_str)
            d.write(self._hinges_str)
            d.write(self._univec_str)
            d.write(self._elem_str)
            d.write(self._loads_str)
            d.write("IEND                0.00            0.00            0.00            0.00")

        print(f'Created an Sesam input deck at "{analysis_dir}"')

    @property
    def _materials_str(self):
        """

        'TDMATER', 'nfield', 'geo_no', 'codnam', 'codtxt', 'name'
        'MISOSEL', 'matno', 'young', 'poiss', 'rho', 'damp', 'alpha', 'iyield', 'yield'

        :return:
        """
        materials = self._gmaterials
        out_str = "".join(
            [self.write_ff("TDMATER", [(4, mat.id, 100 + len(mat.name), 0), (mat.name,)]) for mat in materials]
        )

        out_str += "".join(
            [
                self.write_ff(
                    "MISOSEL",
                    [
                        (mat.id, mat.model.E, mat.model.v, mat.model.rho),
                        (mat.model.zeta, mat.model.alpha, 1, mat.model.sig_y),
                    ],
                )
                for mat in materials
            ]
        )
        return out_str

    @property
    def _sections_str(self):
        """

        'TDSECT', 'nfield', 'geono', 'codnam', 'codtxt', 'set_name'
        'GIORH ', 'geono', 'hz', 'ty', 'bt', 'tt', 'bb', 'tb', 'sfy', 'sfz', 'NLOBYT|', 'NLOBYB|', 'NLOBZ|'

        :return:
        """
        from ada.core.utils import Counter
        from ada.sections import SectionCat

        sec_str = ""
        sec_ids = []
        names_str = ""
        concept_str = ""
        thick_map = dict()
        c = Counter(1)
        g = Counter(1)
        ircon = Counter(1)
        shid = Counter(1)
        bmid = Counter(1)
        for fem_sec in self._gsections:
            assert isinstance(fem_sec, FemSection)
            if fem_sec.type == "beam":
                section = fem_sec.section
                if section not in sec_ids:
                    secid = next(bmid)
                    section.metadata["numid"] = secid
                    names_str += self.write_ff(
                        "TDSECT",
                        [
                            (4, secid, 100 + len(fem_sec.section.name), 0),
                            (fem_sec.section.name,),
                        ],
                    )
                    sec_ids.append(fem_sec.section)
                    if "beam" in fem_sec.metadata.keys():
                        # Give concept relationship based on inputted values
                        beam = fem_sec.metadata["beam"]
                        numel = fem_sec.metadata["numel"]
                        fem_sec.metadata["ircon"] = next(ircon)
                        concept_str += self.write_ff(
                            "TDSCONC",
                            [(4, fem_sec.metadata["ircon"], 100 + len(beam.guid), 0), (beam.guid,)],
                        )
                        concept_str += self.write_ff("SCONCEPT", [(8, next(c), 7, 0), (0, 1, 0, 2)])
                        sconc_ref = next(c)
                        concept_str += self.write_ff("SCONCEPT", [(5, sconc_ref, 2, 4), (1,)])
                        elids = []
                        i = 0
                        elid_bulk = [numel]
                        for el in fem_sec.elset.members:
                            if i == 3:
                                elids.append(tuple(elid_bulk))
                                elid_bulk = []
                                i = -1
                            elid_bulk.append(el.id)
                            i += 1
                        if len(elid_bulk) != 0:
                            elids.append(tuple(elid_bulk))
                            elid_bulk = []

                        mesh_args = [(5 + numel, sconc_ref, 1, 2)] + elids
                        concept_str += self.write_ff("SCONMESH", mesh_args)
                        concept_str += self.write_ff("GUNIVEC", [(next(g), *beam.up)])

                    section.properties.calculate()
                    p = section.properties
                    sec_str += self.write_ff(
                        "GBEAMG",
                        [
                            (secid, 0, p.Ax, p.Ix),
                            (p.Iy, p.Iz, p.Iyz, p.Wxmin),
                            (p.Wymin, p.Wzmin, p.Shary, p.Sharz),
                            (p.Scheny, p.Schenz, p.Sy, p.Sz),
                        ],
                    )

                    if SectionCat.is_i_profile(section.type):
                        sec_str += self.write_ff(
                            "GIORH",
                            [
                                (secid, section.h, section.t_w, section.w_top),
                                (section.t_ftop, section.w_btn, section.t_fbtn, p.Sfy),
                                (p.Sfz,),
                            ],
                        )
                    elif SectionCat.is_hp_profile(section.type):
                        sec_str += self.write_ff(
                            "GLSEC",
                            [
                                (secid, section.h, section.t_w, section.w_btn),
                                (section.t_fbtn, p.Sfy, p.Sfz, 1),
                            ],
                        )
                    elif SectionCat.is_box_profile(section.type):
                        sec_str += self.write_ff(
                            "GBOX",
                            [
                                (secid, section.h, section.t_w, section.t_fbtn),
                                (section.t_ftop, section.w_btn, p.Sfy, p.Sfz),
                            ],
                        )
                    elif SectionCat.is_circular_profile(section.type):
                        sec_str += self.write_ff(
                            "GPIPE",
                            [(secid, section.r - section.wt, section.r, section.wt), (p.Sfy, p.Sfz)],
                        )
                    elif SectionCat.is_flatbar(section.type):
                        sec_str += self.write_ff(
                            "GBARM", [(secid, section.h, section.w_top, section.w_btn), (p.Sfy, p.Sfz)]
                        )
                    else:
                        logging.error(f'Unable to convert "{section}". This will be exported as general section only')

            elif fem_sec.type == "shell":
                if fem_sec.thickness not in thick_map.keys():
                    sh_id = next(shid)
                    thick_map[fem_sec.thickness] = sh_id
                else:
                    sh_id = thick_map[fem_sec.thickness]
                sec_str += self.write_ff("GELTH", [(sh_id, fem_sec.thickness, 5)])
            else:
                raise ValueError(f"Unknown type {fem_sec.type}")
        self._thick_map = thick_map
        return names_str + sec_str + concept_str

    @property
    def _elem_str(self):
        """

        'GELREF1',  ('elno', 'matno', 'addno', 'intno'), ('mintno', 'strano', 'streno', 'strepono'), ('geono', 'fixno',
                    'eccno', 'transno'), 'members|'

        'GELMNT1', 'elnox', 'elno', 'eltyp', 'eltyad', 'nids'

        :return:
        """
        from .reader import SesamReader

        elements = self._gelements

        out_str = "".join(
            [
                self.write_ff(
                    "GELMNT1",
                    [
                        (el.id, el.id, SesamReader.eltype_2_sesam(el.type), 0),
                        ([n.id for n in el.nodes]),
                    ],
                )
                for el in elements
            ]
        )

        def write_elem(el):
            """

            :param el:
            :type el: ada.fem.Elem
            :return: input str for Elem
            """
            fem_sec = el.fem_sec
            assert isinstance(fem_sec, FemSection)
            if fem_sec.type == "beam":
                sec_id = fem_sec.section.metadata["numid"]
            elif fem_sec.type == "shell":
                sec_id = self._thick_map[fem_sec.thickness]
            else:
                raise ValueError(f'Unsupported elem type "{fem_sec.type}"')

            fixno = el.metadata.get("fixno", None)
            transno = el.metadata.get("transno")
            if fixno is None:
                last_tuples = [(sec_id, 0, 0, transno)]
            else:
                h1_fix, h2_fix = fixno
                last_tuples = [(sec_id, -1, 0, transno), (h1_fix, h2_fix)]

            return self.write_ff(
                "GELREF1",
                [
                    (el.id, el.fem_sec.material.id, 0, 0),
                    (0, 0, 0, 0),
                ]
                + last_tuples,
            )

        for el in elements:
            out_str += write_elem(el)

        return out_str

    @property
    def _nodes_str(self):
        """
        GNODE: GNODE NODEX NODENO NDOF ODOF NODEX

        """

        nodes = self._gnodes

        nids = []
        for n in nodes:
            if n.id not in nids:
                nids.append(n.id)
            else:
                raise Exception('Doubly defined node id "{}". TODO: Make necessary code updates'.format(n[0]))
        if len(nodes) == 0:
            return "** No Nodes"
        else:

            out_str = "".join([self.write_ff("GNODE", [(no.id, no.id, 6, 123456)]) for no in nodes])
            out_str += "".join([self.write_ff("GCOORD", [(no.id, no[0], no[1], no[2])]) for no in nodes])
            return out_str

    @property
    def _mass_str(self):
        out_str = ""

        for mass in self._gmass.values():
            for m in mass.fem_set.members:
                if type(mass.mass) in (int, float, np.float64):
                    masses = [mass.mass for _ in range(0, 3)] + [0, 0, 0]
                else:
                    raise NotImplementedError()
                data = (tuple([m.id, 6] + masses[:2]), tuple(masses[2:]))
                out_str += self.write_ff("BNMASS", data)
        return out_str

    @property
    def _bc_str(self):
        out_str = ""
        for bc in self._gbcs:
            for m in bc.fem_set.members:
                dofs = [1 if i in bc.dofs else 0 for i in range(1, 7)]
                data = [tuple([m.id, 6] + dofs[:2]), tuple(dofs[2:])]
                out_str += self.write_ff("BNBCD", data)
        return out_str

    @property
    def _hinges_str(self):
        from ada.core.utils import Counter

        out_str = ""
        h = Counter(1)

        def write_hinge(hinge):
            dofs = [0 if i in hinge else 1 for i in range(1, 7)]
            fix_id = next(h)
            data = [tuple([fix_id, 3, 0, 0]), tuple(dofs[:4]), tuple(dofs[4:])]
            return fix_id, self.write_ff("BELFIX", data)

        for el in self._gelements:
            h1, h2 = el.metadata.get("h1", None), el.metadata.get("h2", None)
            if h2 is None and h1 is None:
                continue
            h1_fix, h2_fix = 0, 0
            if h1 is not None:
                h1_fix, res_str = write_hinge(h1)
                out_str += res_str
            if h2 is not None:
                h2_fix, res_str = write_hinge(h2)
                out_str += res_str
            el.metadata["fixno"] = h1_fix, h2_fix

        return out_str

    @property
    def _univec_str(self):
        from ada.core.utils import Counter

        out_str = ""
        unit_vector = Counter(1)

        def write_local_z(vec):
            transno = next(unit_vector)
            data = [tuple([transno, *vec])]
            return transno, self.write_ff("GUNIVEC", data)

        for el in self._gelements:
            local_z = el.fem_sec.local_z
            transno, res_str = write_local_z(local_z)
            out_str += res_str
            el.metadata["transno"] = transno

        return out_str

    @property
    def _loads_str(self):
        out_str = ""
        for i, l in enumerate(self._gloads):
            assert isinstance(l, Load)
            lid = i + 1
            out_str += self.write_ff("TDLOAD", [(4, lid, 100 + len(l.name), 0), (l.name,)])
            if l.type in ["acc", "grav"]:
                out_str += self.write_ff(
                    "BGRAV",
                    [(lid, 0, 0, 0), tuple([x * l.magnitude for x in l.acc_vector])],
                )
        return out_str

    @staticmethod
    def write_ff(flag, data):
        """
        flag = NCOD
        data = [(int, float, int, float), (float, int)]

        ->> NCOD    INT     FLOAT       INT     FLOAT
                    FLOAT   INT

        :param flag:
        :param data:
        :return:
        """

        def write_data(d):
            if type(d) in (np.float64, float, int, np.uint64, np.int32) and d >= 0:
                return f"  {d:<14.8E}"
            elif type(d) in (np.float64, float, int, np.uint64, np.int32) and d < 0:
                return f" {d:<15.8E}"
            elif type(d) is str:
                return d
            else:
                raise ValueError(f"Unknown input {type(d)} {d}")

        out_str = f"{flag:<8}"
        for row in data:
            v = [write_data(x) for x in row]
            if row == data[-1]:
                out_str += "".join(v) + "\n"
            else:
                out_str += "".join(v) + "\n" + 8 * " "
        return out_str



import os
import pathlib

from ada.config import Settings as _Settings

from .utils import download_to, unzip_it

IFCCONVERT = _Settings.tools_dir / "IfcConvert" / "IfcConvert.exe"


def download_ifc_convert(install_path=IFCCONVERT.parent, ifc_convert_v="v0.6.0-517b819-win64"):
    """
    Download IfcConvert from http://www.ifcopenshell.org/ifcconvert


    :param install_path:
    :param ifc_convert_v:
    :return:
    """
    install_path = pathlib.Path(install_path)

    v_file = install_path / ".version"

    if v_file.exists() is True:
        with open(v_file, "r") as f:
            v = f.read()
            if v == ifc_convert_v:
                print(f'Version "{ifc_convert_v}" is already installed. Skipping downloading')
                return None

    download_path = install_path / "ifc_convert.zip"
    os.makedirs(install_path, exist_ok=True)
    url = f"https://s3.amazonaws.com/ifcopenshell-builds/IfcConvert-{ifc_convert_v}.zip"
    download_to(download_path, url)
    unzip_it(download_path, download_path.parent)

    with open(v_file, "w") as f:
        f.write(ifc_convert_v)



import traceback
from itertools import groupby
from operator import attrgetter

from ada.core.utils import NewLine, bool2text, get_current_user
from ada.fem.io.utils import _folder_prep, get_fem_model_from_assembly

from ..abaqus.writer import (
    AbaqusWriter,
    AbaSection,
    AbaStep,
    _aba_bc_map,
    _valid_aba_bcs,
)
from .execute import run_calculix

main_header_str = """*Heading
** Generated by: {username}
** Assembly For Design and Analysis (ADA) FEM IO (Calculix)
"""


def to_fem(
    assembly,
    name,
    scratch_dir=None,
    description=None,
    execute=False,
    run_ext=False,
    cpus=2,
    gpus=None,
    overwrite=False,
    exit_on_complete=True,
):
    """
    Write a Calculix input file stack

    :param assembly:
    :param name:
    :param scratch_dir:
    :param description:
    :param execute:
    :param run_ext:
    :param cpus:
    :param gpus:
    :param overwrite:
    :param exit_on_complete:
    :return:
    """

    analysis_dir = _folder_prep(scratch_dir, name, overwrite)
    inp_file = (analysis_dir / name).with_suffix(".inp")

    p = get_fem_model_from_assembly(assembly)

    with open(inp_file, "w") as f:
        # Header
        f.write(main_header_str.format(username=get_current_user()))

        # Part level information
        f.write(nodes_str(p.fem.nodes) + "\n")
        f.write(elements_str(p.fem.elements) + "\n")
        f.write(elsets_str(p.fem.elsets) + "\n")
        f.write(elsets_str(assembly.fem.elsets) + "\n")
        f.write(nsets_str(p.fem.nsets) + "\n")
        f.write(nsets_str(assembly.fem.nsets) + "\n")
        f.write(solid_sec_str(p) + "\n")
        f.write(shell_sec_str(p) + "\n")
        f.write(beam_sec_str(p) + "\n")

        # Assembly Level information
        f.write("\n".join([material_str(mat) for mat in p.materials]) + "\n")
        f.write("\n".join([bc_str(x) for x in p.fem.bcs + assembly.fem.bcs]) + "\n")
        f.write(step_str(assembly.fem.steps[0]))
        # f.write(mass_str)
        # f.write(surfaces_str)
        # f.write(constraints_str)
        # f.write(springs_str)

    print(f'Created a Calculix input deck at "{analysis_dir}"')
    # Create run batch files and if execute=True run the analysis
    if execute:
        run_calculix(
            (analysis_dir / name).with_suffix(".inp"),
            cpus=cpus,
            gpus=gpus,
            run_ext=run_ext,
            manifest=assembly.metadata,
            execute=execute,
            exit_on_complete=exit_on_complete,
        )


class CcxSection(AbaSection):
    def __init__(self, fem_sec, fem_writer):
        super().__init__(fem_sec, fem_writer)

    @property
    def section_data(self):
        sec_type = self.fem_sec.section.type
        if "section_type" in self.fem_sec.metadata.keys():
            return self.fem_sec.metadata["section_type"]
        from ada.sections import SectionCat

        if sec_type in SectionCat.circular:
            self.fem_sec.section.properties.calculate()
            return "GENERAL"
        elif sec_type in SectionCat.igirders or sec_type in SectionCat.iprofiles:
            self.fem_sec.section.properties.calculate()
            return "GENERAL"
        elif sec_type in SectionCat.box:
            return "BOX"
        elif sec_type in SectionCat.general:
            return "GENERAL"
        elif sec_type in SectionCat.tubular:
            self.fem_sec.section.properties.calculate()
            return "PIPE"
        elif sec_type in SectionCat.angular:
            self.fem_sec.section.properties.calculate()
            return "GENERAL"
        else:
            raise Exception(f'Section type "{sec_type}" is not added to AbaBeam yet.\n{traceback.format_exc()}')

    @property
    def beam_str(self):
        """
        BOX, CIRC, HEX, I, L, PIPE, RECT, THICK PIPE, and TRAPEZOID sections
        https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-beamsection.htm


        General Section
        https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-beamgeneralsection.htm#simakey-r-beamgeneralsection__simakey-r-beamgeneralsection-s-datadesc1


        Comment regarding Rotary Inertia and Explicit analysis
        https://abaqus-docs.mit.edu/2017/English/SIMACAEELMRefMap/simaelm-c-beamsectionbehavior.htm#hj-top

        """
        top_line = f"** Section: {self.fem_sec.elset.name}  Profile: {self.fem_sec.elset.name}"
        n1 = ", ".join(str(x) for x in self.fem_sec.local_y)
        ass = self.fem_sec.parent.parent.get_assembly()
        rotary_str = ""
        if len(ass.fem.steps) > 0:
            initial_step = ass.fem.steps[0]
            if initial_step.type == "explicit":
                rotary_str = ", ROTARY INERTIA=ISOTROPIC"

        second_line = (
            f"*Beam Section, elset={self.fem_sec.elset.name}, material={self.fem_sec.material.name}, "
            f"section={self.section_data}{rotary_str}"
        )

        if self.section_data != "GENERAL":
            return f"""{top_line}
{second_line}
 {self.props}"""
        elif self.section_data == "PIPE":
            return f"{self.fem_sec.section.r}, {self.fem_sec.section.wt}\n {n1}"
        else:
            return f"""{top_line}
*Beam Section, elset={self.fem_sec.elset.name}, material={self.fem_sec.material.name},  section=GENERAL{rotary_str}
 {self.props}"""


def step_str(step):
    """

    :param step:
    :type step: ada.fem.Step
    :return:
    """

    bcstr = "\n".join([bc_str(bc) for bc in step.bcs.values()]) if len(step.bcs) > 0 else "** No BCs"
    lstr = "\n".join([load_str(l) for l in step.loads]) if len(step.loads) > 0 else "** No Loads"

    int_str = (
        "\n".join([interactions_str(interact) for interact in step.interactions.values()])
        if len(step.interactions.values()) > 0
        else "** No Interactions"
    )

    nodal = []
    elem = []
    for fi in step.field_outputs:
        nodal += fi.nodal
        elem += fi.element
    nodal_str = "*node file\n" + ", ".join(nodal) if len(nodal) > 0 else "** No nodal output"
    elem_str = "*el file\n" + ", ".join(elem) if len(elem) > 0 else "** No elem output"

    return f"""**
** STEP: {step.name}
**
*Step, nlgeom={bool2text(step.nl_geom)}, inc={step.total_incr}
*Static
 {step.init_incr}, {step.total_time}, {step.min_incr}, {step.max_incr}
**
** BOUNDARY CONDITIONS
**
{bcstr}
**
** LOADS
**
{lstr}
**
** INTERACTIONS
**
{int_str}
**
** OUTPUT REQUESTS
**
{nodal_str}
{elem_str}
*End Step"""


def nodes_str(fem_nodes):
    f = "{nid:>7}, {x:>13}, {y:>13}, {z:>13}"
    return (
        "*NODE\n"
        + "\n".join(
            [f.format(nid=no.id, x=no[0], y=no[1], z=no[2]) for no in sorted(fem_nodes, key=attrgetter("id"))]
        ).rstrip()
        if len(fem_nodes) > 0
        else "** No Nodes"
    )


def elements_str(fem_elements):
    """

    :param fem_elements:
    :return:
    """
    from ..abaqus.writer import elwriter

    return (
        "".join(
            filter(
                None,
                [elwriter(x, elements) for x, elements in groupby(fem_elements, key=attrgetter("type", "elset"))],
            )
        ).rstrip()
        if len(fem_elements) > 0
        else "** No elements"
    )


def gen_set_str(fem_set):
    if len(fem_set.members) == 0:
        if "generate" in fem_set.metadata.keys():
            if fem_set.metadata["generate"] is False:
                raise ValueError(f'set "{fem_set.name}" is empty. Please check your input')
        else:
            raise ValueError("No Members are found")

    generate = fem_set.metadata.get("generate", False)
    internal = fem_set.metadata.get("internal", False)
    newline = NewLine(15)

    el_str = "*Elset, elset" if fem_set.type == "elset" else "*Nset, nset"

    el_instances = dict()

    for p, mem in groupby(fem_set.members, key=attrgetter("parent")):
        el_instances[p.name] = list(mem)

    # for mem in self.members:
    #     if mem.parent.name not in el_instances.keys():
    #         el_instances[mem.parent.name] = []
    #     if mem not in el_instances[mem.parent.name]:
    #         el_instances[mem.parent.name].append(mem)

    set_str = ""
    for elinst, members in el_instances.items():
        el_root = f"{el_str}={fem_set.name}"
        if internal is True and type(fem_set._fem_writer) in (
            AbaqusWriter,
            AbaStep,
        ):
            el_root += "" if "," in el_str[-2] else ", "
            el_root += "internal"

        if generate:
            assert len(fem_set.metadata["gen_mem"]) == 3
            el_root += "" if "," in el_root[-2] else ", "
            set_str += (
                el_root + "generate\n {},  {},   {}" "".format(*[no for no in fem_set.metadata["gen_mem"]]) + "\n"
            )
        else:
            set_str += el_root + "\n " + " ".join([f"{no.id}," + next(newline) for no in members]).rstrip()[:-1] + "\n"
    return set_str.rstrip()


def elsets_str(fem_elsets):
    if len(fem_elsets) > 0:
        return "\n".join([gen_set_str(el) for el in fem_elsets.values()]).rstrip()
    else:
        return "** No element sets"


def nsets_str(fem_nsets):
    return (
        "\n".join([gen_set_str(no) for no in fem_nsets.values()]).rstrip() if len(fem_nsets) > 0 else "** No node sets"
    )


def solid_sec_str(part):
    solid_secs = [AbaSection(sec, part).str for sec in part.fem.sections.solids]
    return "\n".join(solid_secs).rstrip() if len(solid_secs) > 0 else "** No solid sections"


def shell_sec_str(part):
    shell_secs = [AbaSection(sec, part).str for sec in part.fem.sections.shells]
    return "\n".join(shell_secs).rstrip() if len(shell_secs) > 0 else "** No shell sections"


def beam_sec_str(part):
    beam_secs = [CcxSection(sec, part).str for sec in part.fem.sections.beams]
    return "\n".join(beam_secs).rstrip() if len(beam_secs) > 0 else "** No beam sections"


def material_str(material):
    if "aba_inp" in material.metadata.keys():
        return material.metadata["aba_inp"]
    if "rayleigh_damping" in material.metadata.keys():
        alpha, beta = material.metadata["rayleigh_damping"]
    else:
        alpha, beta = None, None

    no_compression = material._metadata["no_compression"] if "no_compression" in material._metadata.keys() else False
    compr_str = "\n*No Compression" if no_compression is True else ""

    if material.model.eps_p is not None and len(material.model.eps_p) != 0:
        pl_str = "\n*Plastic\n"
        pl_str += "\n".join(
            ["{x:>12.5E}, {y:>10}".format(x=x, y=y) for x, y in zip(material.model.sig_p, material.model.eps_p)]
        )
    else:
        pl_str = ""

    if alpha is not None and beta is not None:
        d_str = "\n*Damping, alpha={alpha}, beta={beta}".format(alpha=material.model.alpha, beta=material.model.beta)
    else:
        d_str = ""

    if material.model.zeta is not None and material.model.zeta != 0.0:
        exp_str = "\n*Expansion\n {zeta}".format(zeta=material.model.zeta)
    else:
        exp_str = ""

    return f"""*Material, name={material.name}
*Elastic
 {material.model.E:.6E},  {material.model.v}{compr_str}
*Density
 {material.model.rho},{exp_str}{d_str}{pl_str}"""


def bc_str(bc):
    """

    :param bc:
    :type bc: ada.fem.Bc
    :return:
    """
    ampl_ref_str = "" if bc.amplitude_name is None else ", amplitude=" + bc.amplitude_name

    if bc.type in _valid_aba_bcs:
        aba_type = bc.type
    else:
        aba_type = _aba_bc_map[bc.type]

    dofs_str = ""
    for dof, magn in zip(bc.dofs, bc.magnitudes):
        if dof is None:
            continue
        # magn_str = f", {magn:.4f}" if magn is not None else ""

        if bc.type in ["connector displacement", "connector velocity"] or type(dof) is str:
            inst_name = bc.fem_set.name
            dofs_str += f" {inst_name}, {dof}\n"
        else:
            inst_name = bc.fem_set.name
            dofs_str += f" {inst_name}, {dof}\n"

    dofs_str = dofs_str.rstrip()

    if bc.type == "connector displacement":
        bcstr = "*Connector Motion"
        add_str = ", type=DISPLACEMENT"
    elif bc.type == "connector velocity":
        bcstr = "*Connector Motion"
        add_str = ", type=VELOCITY"
    else:
        bcstr = "*Boundary"
        add_str = ""

    return f"""** Name: {bc.name} Type: {aba_type}
{bcstr}{ampl_ref_str}{add_str}
{dofs_str}"""


def load_str(load):
    """

    :param load:
    :type load: ada.fem.Load
    :return:
    """
    dof = [0, 0, 1] if load.dof is None else load.dof
    if load.fem_set is None:
        raise ValueError("Calculix does not accept Loads without reference to a fem_set")

    fem_set = load.fem_set.name
    return f"""** Name: gravity   Type: Gravity
*Dload
{fem_set}, GRAV, {load.magnitude}, {', '.join([str(x) for x in dof[:3]])}"""


def surface_str(surface):
    """

    :param surface:
    :type surface: ada.fem.Surface
    :return:
    """
    top_line = f"*Surface, type={surface.type}, name={surface.name}"
    id_refs_str = "\n".join([f"{m[0]}, {m[1]}" for m in surface.id_refs]).strip()
    if surface.id_refs is None:
        if surface.type == "NODE":
            add_str = surface.weight_factor
        else:
            add_str = surface.face_id_label
        if surface.fem_set.name in surface.parent.elsets.keys():
            return f"{top_line}\n{surface.fem_set.name}, {add_str}"
        else:
            return f"""{top_line}
{surface.fem_set.name}, {add_str}"""
    else:
        return f"""{top_line}
{id_refs_str}"""


def interactions_str(interaction):
    """

    :param interaction:
    :type interaction: ada.fem.Interaction
    :return:
    """
    from ada.fem import Step

    if interaction.type == "SURFACE":
        adjust_par = interaction.metadata.get("adjust", None)
        geometric_correction = interaction.metadata.get("geometric_correction", None)
        small_sliding = interaction.metadata.get("small_sliding", None)

        stpstr = f"*Contact Pair, interaction={interaction.interaction_property.name}"

        if small_sliding is not None:
            stpstr += f", {small_sliding}"

        if type(interaction.parent) is Step:
            step = interaction.parent
            assert isinstance(step, Step)
            stpstr += "" if "explicit" in step.type else f", type={interaction.surface_type}"
        else:
            stpstr += f", type={interaction.surface_type}"

        if interaction.constraint is not None:
            stpstr += f", mechanical constraint={interaction.constraint}"

        if adjust_par is not None:
            stpstr += f", adjust={adjust_par}" if adjust_par is not None else ""

        if geometric_correction is not None:
            stpstr += f", geometric correction={geometric_correction}"

        stpstr += f"\n{interaction.surf1.name}, {interaction.surf2.name}"
    else:
        raise NotImplementedError(f'type "{interaction.type}"')

    return f"""**
** Interaction: {interaction.name}
{stpstr}"""
